import argparse
import textwrap
from functools import partial
from glob import glob
from multiprocessing import Pool
from pathlib import Path
from typing import Literal, Optional

import matplotlib.font_manager as fm
import matplotlib.image as mpimage
import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle
from matplotlib.text import Annotation
from matplotlib.transforms import Bbox
from tqdm import tqdm


from shithappens.card import Card
from shithappens.cli.utils import verify_input_dir
from shithappens.utils import merge_pdfs, slugify


def text_with_wrap_autofit(
    ax: plt.Axes,
    txt: str,
    xy_size: tuple[float, float],
    width: float,
    height: float,
    *,
    min_font_size=None,
    bleed: Optional[float] = None,
    pad: Optional[float] = None,
    show_rect: bool = False,
    **kwargs,
):
    """Automatically fits the text to some axes.

    Args:
        ax: axes to put the text on.
        txt: text to display.
        xy: location to place the text.
        width: width of the text box in fractions.
        height: height of the text box in fractions.
        min_font_size: minimum acceptable font size.
        bleed: bleed of the figure.
        pad: padding of the box.
        **kwargs: keyword arguments passed to Axes.annotate.

    Returns:
        text artist.
    """

    #  Different alignments give different bottom left and top right anchors.
    x, y = xy_size
    if bleed is None:
        bleed = 0
    if pad:
        bleed += pad
        x -= 2 * pad
        y -= 2 * pad

    if show_rect:
        alpha = 0.3
    else:
        alpha = 0

    rect = Rectangle(
        (bleed + (1 - width) * x, bleed + (1 - height) * y),
        width * x,
        height * y,
        alpha=alpha,
    )
    ax.add_patch(rect)

    # Get transformation to go from display to data-coordinates.
    inv_data = ax.transData.inverted()

    fig: Figure = ax.get_figure()
    dpi = fig.dpi
    rect_height_inch = rect.get_height() / dpi

    # Initial fontsize according to the height of boxes
    fontsize = rect_height_inch * 72

    wrap_lines = 1
    xy = (bleed + 0.5 * x, bleed + 0.95 * y)
    while True:
        wrapped_txt = "\n".join(
            textwrap.wrap(txt, width=len(txt) // wrap_lines, break_long_words=False)
        )

        # For dramatic effect, place text after ellipsis on newline.
        wrapped_txt = wrapped_txt.replace("... ", "...\n")
        wrapped_txt = wrapped_txt.replace("… ", "...\n")
        text: Annotation = ax.annotate(wrapped_txt, xy, **kwargs)
        text.set_fontsize(fontsize)

        # Adjust the fontsize according to the box size.
        bbox: Bbox = text.get_window_extent()
        inv_text_bbox = inv_data.transform(bbox)
        width_text = inv_text_bbox[1][0] - inv_text_bbox[0][0]
        adjusted_size = fontsize * rect.get_width() / width_text
        if min_font_size is None or adjusted_size >= min_font_size:
            break
        text.remove()
        wrap_lines += 1
    text.set_fontsize(adjusted_size)

    return text


class ShitHappensArgs(argparse.Namespace):
    input_dir: str
    name: str
    merge: bool
    rank: bool
    side: Literal["front", "back", "both"]
    format: Literal["pdf", "png"]
    workers: int
    chunks: int


def parse_excel(input_path: Path, desc_col: int, misery_index_col: int) -> pd.DataFrame:
    """Parse an Excel file.
    It must have two colums: descriptions along with their misery index.

    Args:
        intput_path: path of the Excel file
        desc_col: description column index
        misery_index_col: misery index column index

    Returns:
        Pandas DataFrame with index, description, and misery index.
    """
    try:
        df = pd.read_excel(
            input_path, usecols=[desc_col, misery_index_col], engine="openpyxl"
        )
    except Exception:
        print(f"{input_path} is not an excel file.")
        exit()

    return df


def plot_crop_marks(ax: Axes, bleed: float, factor: float = 0.6):
    """Plots crop marks on the given axis.
    The crop marks will mark the bleed. The crop mark size is adjustable with the factor.
    """
    crop_mark_len = factor * bleed
    fig = ax.get_figure()
    bbox: Bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    x_size, y_size = bbox.width, bbox.height

    # h, v - horizontal, vertical
    # u, d - up, down
    # l, r - left, right
    hul = (y_size - bleed, 0, crop_mark_len)
    hdl = (bleed, 0, crop_mark_len)
    hur = (y_size - bleed, x_size - crop_mark_len, x_size)
    hdr = (bleed, x_size - crop_mark_len, x_size)
    vul = (bleed, y_size - crop_mark_len, y_size)
    vdl = (bleed, 0, crop_mark_len)
    vur = (x_size - bleed, y_size - crop_mark_len, y_size)
    vdr = (x_size - bleed, 0, crop_mark_len)

    cropmarkstyle = {"color": "white", "linewidth": 1}

    for horizontal_mark in [hul, hdl, hur, hdr]:
        ax.hlines(*horizontal_mark, **cropmarkstyle)
    for vertical_mark in [vul, vdl, vur, vdr]:
        ax.vlines(*vertical_mark, **cropmarkstyle)


def plot_card_front(card: Card) -> Figure:
    # To be able to convert between centimeters and inches.
    cm_per_inch = 2.54

    # 62x88 mm for typical playing cards.
    x_size = 6.2 / cm_per_inch  # cm front and back
    y_size = 8.8 / cm_per_inch  # cm top to bottom

    # Add margin on all sides.
    bleed = 0.5 / cm_per_inch  # cm
    pad = 0.3 / cm_per_inch

    x_total = x_size + 2 * bleed
    y_total = y_size + 2 * bleed
    xy_size = (x_total, y_total)

    fig, ax = plt.subplots()

    fig.set_size_inches(*xy_size)
    fig.set_facecolor("black")
    fig.subplots_adjust(left=0, bottom=0, right=1, top=1)

    ax.axis("off")

    ax.set_xlim(0, x_total)
    ax.set_ylim(0, y_total)

    opensans_path = Path(__file__).parent.resolve() / Path(
        "opensans/fonts/ttf/OpenSans-ExtraBold.ttf"
    )
    prop = fm.FontProperties(fname=opensans_path)

    text_kwargs = dict(wrap=True, horizontalalignment="center", fontproperties=prop)

    # Front.
    text_with_wrap_autofit(
        ax,
        card.desc.upper(),
        (x_size, y_size),
        1,
        0.4,
        **text_kwargs,
        bleed=bleed,
        pad=pad,
        min_font_size=11,
        va="top",
        weight="extra bold",
        color="yellow",
    )

    mi_desc = "misery index"
    ax.text(
        x_total / 2,
        1.3 * y_size / 8 + bleed,
        mi_desc.upper(),
        **text_kwargs,
        color="yellow",
        fontsize=13,
        weight="extra bold",
        verticalalignment="center",
    )

    ax.text(
        x_total / 2,
        0.05 * y_size + bleed,
        card.misery_index if ".5" in str(card.misery_index) else int(card.misery_index),
        **text_kwargs,
        color="black",
        fontsize=23,
        weight="extra bold",
        verticalalignment="center",
    )

    mi_block = Rectangle(
        (bleed + x_size / 4, 0), x_size / 2, y_size / 8 + bleed, fc="yellow"
    )
    ax.add_patch(mi_block)

    plot_crop_marks(ax, bleed)

    plt.close(fig)

    return fig


def plot_card_back(card: Card, input_dir: Path) -> Figure:
    # To be able to convert between centimeters and inches.
    cm_per_inch = 2.54

    # 62x88 mm for typical playing cards.
    x_size = 6.2 / cm_per_inch  # cm front and back
    y_size = 8.8 / cm_per_inch  # cm top to bottom

    # Add margin on all sides.
    bleed = 0.5 / cm_per_inch  # cm

    x_total = x_size + 2 * bleed
    y_total = y_size + 2 * bleed
    xy_size = (x_total, y_total)

    fig, ax = plt.subplots()

    fig.set_size_inches(*xy_size)
    fig.set_facecolor("black")
    fig.subplots_adjust(left=0, bottom=0, right=1, top=1)

    ax.axis("off")

    parent_dir = Path(__file__).parent.resolve()

    opensans_regular_path = parent_dir / Path("opensans/fonts/ttf/OpenSans-Regular.ttf")
    prop_regular = fm.FontProperties(fname=opensans_regular_path)

    text_kwargs = dict(
        wrap=True, horizontalalignment="center", fontproperties=prop_regular
    )

    game_name = "Shit Happens"
    expansion_text = "edition"
    expansion_text_full = card.expansion_name + " " + expansion_text

    ax.text(
        x_size / 2 + bleed,
        0.9 * y_size + bleed,
        game_name.upper(),
        **text_kwargs,
        color="yellow",
        fontsize=20,
        weight="regular",
        verticalalignment="center",
    )

    opensans_light_path = parent_dir / Path("opensans/fonts/ttf/OpenSans-Regular.ttf")
    prop_light = fm.FontProperties(fname=opensans_light_path)

    text_kwargs = dict(
        wrap=True, horizontalalignment="center", fontproperties=prop_light
    )

    ax.text(
        x_size / 2 + bleed,
        0.83 * y_size + bleed,
        expansion_text_full.upper(),
        **text_kwargs,
        color="yellow",
        fontsize=14,
        fontstyle="italic",
        weight="ultralight",
        verticalalignment="center",
    )

    # Expansion logo

    try:
        expansion_logo_path = glob(f"{input_dir}/**/expansion-logo.*")[0]
    except KeyError:
        expansion_logo_path = str(parent_dir / Path("images/expansion-logo.png"))

    expansion_logo = mpimage.imread(expansion_logo_path)

    expansion_logoax = fig.add_axes([0.2, 0.1, 0.6, 0.6])
    expansion_logoax.imshow(
        expansion_logo,
    )
    expansion_logoax.axis("off")

    plot_crop_marks(ax, bleed)

    ax.set_xlim(0, x_total)
    ax.set_ylim(0, y_total)

    plt.close(fig)

    return fig


def save_card(
    card: Card,
    output_dir: Path,
    side: Literal["front", "back"],
    dpi: int = 300,
    format: str = "pdf",
) -> None:
    side_fn = "front" if side == "front" else "back"

    output_dir = output_dir / side_fn

    output_dir.mkdir(parents=True, exist_ok=True)

    fn = f"{card.misery_index}-{card.desc}"
    fn = slugify(fn)
    save_fn = (output_dir / fn).with_suffix("." + format)

    if side == "front":
        card.fig_front.savefig(
            str(save_fn),
            format=save_fn.suffix[1:],
            pad_inches=0,
            dpi=dpi,
            transparent=False,
        )
    elif side == "back":
        card.fig_back.savefig(
            str(save_fn),
            format=save_fn.suffix[1:],
            pad_inches=0,
            dpi=dpi,
            transparent=False,
        )


def create_card(
    row, expansion_name, input_dir, output_dir, side, ext: Literal["pdf", "png"]
):
    card = Card(row[1]["desc"], row[1]["misery_index"], expansion_name)

    if side == "front" or side == "both":
        card.fig_front = plot_card_front(card)
        save_card(card, output_dir, "front", format=ext)

    if side == "back" or side == "both":
        card.fig_back = plot_card_back(card, input_dir)
        save_card(card, output_dir, "back", format=ext)


def create_cards(
    df: pd.DataFrame,
    expansion_name: str,
    input_dir: Path,
    output_dir: Path,
    merge: bool,
    side: Literal["front", "back", "both"],
    ext: Literal["pdf", "png"],
    workers: int,
    chunks: int,
) -> None:
    nmax = df.shape[0]
    chunksize = nmax // chunks
    create_card_par = partial(
        create_card,
        expansion_name=expansion_name,
        input_dir=input_dir,
        output_dir=output_dir,
        side=side,
        ext=ext,
    )
    desc = "Plotting cards"
    if chunksize:
        with Pool(workers) as p:
            list(
                tqdm(
                    p.imap_unordered(create_card_par, df.iterrows(), chunksize),
                    total=nmax,
                    desc=desc,
                )
            )
    else:
        list(tqdm(map(create_card_par, df.iterrows()), total=nmax, desc=desc))

    if merge:
        if side == "front" or side == "both":
            merge_pdfs(output_dir / "front")
        if side == "back" or side == "both":
            merge_pdfs(output_dir / "back")


def main(**args) -> None:
    input_dir = Path(args["input_dir"])
    xlsx_path, output_dir = verify_input_dir(input_dir)

    if False:
        pass
    # if args["rank"]:
    #     from shithappens.sort_situations import sort

    #     sort(xlsx_path)
    else:
        if args["name"]:
            expansion_name = args["name"]
        else:
            expansion_name = input_dir.stem
            print(
                "Argument -n/--name not given. "
                f"Expansion name inferred to be {expansion_name}."
            )

        df = parse_excel(xlsx_path, 0, 1)

        create_cards(
            df,
            expansion_name,
            input_dir,
            output_dir,
            args["merge"],
            args["side"],
            args["format"],
            args["workers"],
            args["chunks"],
        )


def main_cli(**kwargs):
    try:
        main(**kwargs)
    except KeyboardInterrupt:
        print("Interrupted.")


if __name__ == "__main__":
    main()



from pathlib import Path
from glob import glob
import click


def verify_input_dir(input_dir: Path):
    while True:
        if input_dir.exists():
            break
        input_dir = Path(
            input(
                f"Input directory {input_dir} does not exist. "
                "Please specify an existing input directory.\n"
            )
        )

    output_dir = input_dir / "outputs"
    print(f"Reading files from {input_dir}.")
    print(f"Output files in {output_dir}.")

    xlsx_paths = glob(f"{input_dir / '*.xlsx'}")
    xlsx_paths_num = len(xlsx_paths)
    if not xlsx_paths_num:
        print("Please provide an Excel file in {input_dir}.")
        exit(1)
    elif xlsx_paths_num > 1:
        while True:
            print("\nMore than one input file found.")
            for i, xlsx_path in enumerate(xlsx_paths, 1):
                print(f"[{i}] {xlsx_path}")
            try:
                xlsx_index = int(click.getchar())
                xlsx_path = Path(xlsx_paths[xlsx_index - 1])
            except (ValueError, IndexError):
                continue
            else:
                break

    else:
        xlsx_path = Path(xlsx_paths[0])

    return xlsx_path, output_dir



import re
import unicodedata
from glob import glob
from pathlib import Path
from typing import Optional


def sort_pdfs_by_mi(fn: str):
    """Key for sorted() to sort filenames by leading number.
    The number must be separated with '-'.
    """
    try:
        return float(Path(fn).stem.split("-")[0])
    except ValueError:
        return -1


def merge_pdfs(input_dir: Path, output_dir: Optional[Path] = None):
    """Merges all the pdf files in current directory.
    Source: https://stackoverflow.com/a/47356404/8797886.
    """
    from PyPDF2 import PdfMerger

    merger = PdfMerger()
    allpdfs = [
        a
        for a in sorted(
            glob(str(input_dir / "*.pdf")),
            key=sort_pdfs_by_mi,
        )
        if "merged" not in a
    ]
    for pdf in allpdfs:
        merger.append(pdf)

    if output_dir is None:
        output_dir = input_dir

    with open(output_dir / "merged.pdf", "wb") as merged_pdf:
        merger.write(merged_pdf)


def slugify(value, allow_unicode=False):
    """
    Convert to ASCII if 'allow_unicode' is False. Convert spaces or repeated
    dashes to single dashes. Remove characters that aren't alphanumerics,
    underscores, or hyphens. Convert to lowercase. Also strip leading and
    trailing whitespace, dashes, and underscores.
    Source: https://github.com/django/django/blob/0b78ac3fc7bd9f0c57518d0c1a153582318edd59/django/utils/text.py#L420.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = (
            unicodedata.normalize("NFKD", value)
            .encode("ascii", "ignore")
            .decode("ascii")
        )
    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")



from dataclasses import dataclass, field

from matplotlib.figure import Figure


@dataclass
class Card:
    desc: str
    misery_index: float
    expansion_name: str
    fig_front: Figure = field(init=False)
    fig_back: Figure = field(init=False)



import click

from shithappens.create_cards import main_cli as create_cli
from shithappens.sort_situations import main_cli as sort_cli


@click.group(help="Create custom Shit Happens expansion playing cards.")
def cli():
    pass


@cli.command(help="Create cards.")
@click.argument("input_dir")
@click.option(
    "-n",
    "--name",
    "name",
    help="Expansion name. If no name is specified, infers name from input_dir.",
)
@click.option("-m", "--merge", "merge", is_flag=True, help="Merge output.")
@click.option(
    "-s",
    "--side",
    "side",
    type=click.Choice(["front", "back", "both"], case_sensitive=False),
    default="both",
    help="Side(s) to generate.",
)
@click.option(
    "-f",
    "--format",
    "format",
    type=click.Choice(["pdf", "png"], case_sensitive=False),
    default="pdf",
    help="Output format.",
)
@click.option(
    "-w", "--workers", "workers", type=int, default=4, help="Number of workers."
)
@click.option(
    "-c",
    "--chunks",
    "chunks",
    type=int,
    default=30,
    help="Number of chunks for the workers to process",
)
def create(**kwargs):
    create_cli(**kwargs)


@cli.command(help="Rank situation.")
@click.argument("input_dir")
@click.option(
    "-s",
    "--strategy",
    "strategy",
    type=click.Choice(["swiss", "round-robin"], case_sensitive=False),
    default="swiss",
    help="Ranking strategy.",
)
@click.option(
    "-r",
    "--rounds",
    "rounds",
    type=int,
    default=9,
    help="The number of rounds to use with the swiss strategy.",
)
@click.option(
    "-p",
    "--prescore",
    "prescore",
    type=int,
    default=10,
    help="The number of groups to prescore.",
)
def rank(**kwargs):
    sort_cli(**kwargs)



import itertools
from pathlib import Path
from typing import Literal

import click
import pandas as pd
from tqdm import tqdm

UP = "\x1b[3A"
ERASE = "\x1b[2A"


def parse_excel(input_path: Path, desc_col: int, score_col: int = 2) -> pd.DataFrame:
    """Parse an Excel file.
    It must have two colums: descriptions along with their misery index.

    Args:
        intput_path: path of the Excel file
        desc_col: description column index
        misery_index_col: misery index column index

    Returns:
        Pandas DataFrame with index, description, and misery index.
    """
    try:
        df = pd.read_excel(input_path, usecols=[desc_col, score_col], engine="openpyxl")
    except Exception:
        print(f"{input_path} does not contain any Excel files.")
        exit()

    return df


def prompt_question(df: pd.DataFrame, situations, round=None):
    tqdm.write(f"{UP}")
    for i, situation in enumerate(situations, 1):
        tqdm.write(f"[{i}] #{situation} {df.loc[situation]['desc']}")
    try:
        most_miserable_idx = int(click.getchar())
        if most_miserable_idx == 0:
            raise ValueError("0 is not a valid choice.")
        df.loc[situations[most_miserable_idx - 1], "score"] += 1
        if round is not None:
            df.loc[situations[0], f"round_{round}"] = situations[1]
            df.loc[situations[1], f"round_{round}"] = situations[0]
    except (ValueError, IndexError):
        prompt_question(df, situations)


def save(xlsx_path, df):
    if "sorted" in xlsx_path.stem:
        in_place = True
    else:
        in_place = False

    if in_place:
        output_file = xlsx_path
    else:
        output_file = xlsx_path.with_name(xlsx_path.stem + "-sorted.xlsx")
    df.to_excel(output_file, index=False)

    if len(df["score"].unique()) != len(df):
        print(
            "Some items have equal ranking score. "
            f"Please check the output file ({output_file})."
        )
    print("Manually assign a misery-index to situations.")


def sort(
    xlsx_path,
    output_dir,
    strategy: Literal["swiss", "round-robin"] = "swiss",
    rounds=9,
    prescore=10,
):
    df = parse_excel(xlsx_path, 0)

    if strategy == "swiss":
        print("\n")
        if prescore == 10:
            print("Rate the situation 1-10. (Press 0 for rate 10.)")
        else:
            print("Rate this situation 1-{}".format(prescore))
        try:
            df["score"]
        except KeyError:
            df["score"] = 0
        # First, give items scores 1-9 and 0, with 0=10.
        with tqdm(
            df.iterrows(),
            desc="Prescore",
            total=len(df),
            maxinterval=1,
            unit="situation",
        ) as prog_iter:
            for i, situation in prog_iter:
                tqdm.write(ERASE)
                tqdm.write(situation["desc"])
                while True:
                    try:
                        score_group = int(click.getchar())
                    except ValueError:
                        continue
                    else:
                        break
                df.loc[i, "score"] += score_group if score_group > 0 else 10
        df = df.sort_values(by="score")

        splits = []
        num_groups = 10
        step = int(len(df) / (num_groups - 1))

        # Force step to be even.
        if step % 2 != 0:
            step -= 1
        i = 0
        while i < len(df):
            splits.append(i)
            i += step
        if splits[-1] != len(df) - 1:
            splits.append(len(df) - 1)

        print("\n\n\n")
        print("Which situation is worse?")

        for round_num in tqdm(
            range(rounds), desc="Round", total=rounds, unit="round", maxinterval=1
        ):
            df[f"round_{round_num}"] = None
            for split_begin, split_end in tqdm(
                zip(splits, splits[1:]),
                total=len(splits),
                unit="split",
                leave=False,
                position=1,
                maxinterval=1,
                desc="Split",
            ):
                range_list = range(split_begin, split_end)
                index_list_temp = zip(range_list, range_list[::-1])
                index_list = []
                for item in index_list_temp:
                    if item[0] < item[1]:
                        index_list.append(item)
                for idx1, idx2 in tqdm(
                    index_list,
                    total=len(index_list),
                    maxinterval=1,
                    leave=False,
                    position=2,
                    desc="Comparison",
                    unit="comparison",
                ):
                    df_idx1, df_idx2 = df.index[idx1], df.index[idx2]
                    df_temp_idx2 = df_idx2
                    # No rematches.
                    while True:
                        if idx1 in [
                            df.loc[df_temp_idx2, f"round_{round_n}"]
                            for round_n in range(round_num + 1)
                        ]:
                            df_temp_idx2 -= 1
                        else:
                            break
                    if idx1 >= idx2:
                        with open("test.txt", "a") as f:
                            f.write(f"{idx1}, {idx2}\n")
                        break
                    prompt_question(df, [df_idx1, df_temp_idx2], round_num)
            df = df.sort_values(by="score")

            save(xlsx_path, df)

    elif strategy == "round-robin":
        try:
            continue_from = df["score"].sum()
            print(f"Continuing from situation {continue_from}")
        except KeyError:
            df["misery_index"] = 0
            df["score"] = 0
            continue_from = 0

        combinations = list(itertools.combinations(df.index, 2))[continue_from:]

        print("\nWhich situation is most miserable?\n\n")

        try:
            with tqdm(
                combinations,
                total=len(combinations),
                initial=continue_from,
                smoothing=0,
            ) as progress_iterator:
                for combination in progress_iterator:
                    prompt_question(df, combination)
        except KeyboardInterrupt:
            print("Do you want to save your progress? [y]/n")
            save_progress = click.getchar()
            if not save_progress:
                save_progress = True
            elif save_progress == "n":
                save_progress = False

            if save_progress:
                save(xlsx_path, df)
        else:
            save(xlsx_path, df)


def main_cli(**kwargs):
    try:
        from shithappens.cli.utils import verify_input_dir

        xlsx_path, output_dir = verify_input_dir(Path(kwargs["input_dir"]))
        sort(
            xlsx_path,
            output_dir,
            kwargs["strategy"],
            kwargs["rounds"],
            kwargs["prescore"],
        )
    except KeyboardInterrupt:
        print("Interrupted.")
